var ref = require('react');
var useContext = ref.useContext;
var useEffect = ref.useEffect;
var useState = ref.useState;
var useRef = ref.useRef;
var ref$1 = require('../react');
var AtomContext = ref$1.AtomContext;
var differ = ref$1.differ;
var raf = require('../raf')

var isServer = typeof navigator === 'undefined'

function useAtom (selector, options) {
  if ( options === void 0 ) options = {};

  selector = selector || (function (state) { return state; })
  var sync = options.sync; if ( sync === void 0 ) sync = false;
  var pure = options.pure; if ( pure === void 0 ) pure = true;
  var observe = options.observe; if ( observe === void 0 ) observe = !isServer;
  var schedule = sync ? function (fn) { return function () { return fn(); }; } : raf
  var ref$1 = useContext(AtomContext);
  var atom = ref$1.atom;
  assert(atom, 'No atom found in context, did you forget to wrap your app in <Provider atom={atom} />?')
  var ref$2 = useState({});
  var triggerRerender = ref$2[1];
  var ref = useRef({ mappedProps: null, atom: null })

  ref.current.atom = atom
  ref.current.mappedProps = selector(atom.get())

  // keep track of rendering order
  // this is important for correctness – parent must rerender first
  // and performance – parent rerendering children should cancel
  // children's scheduled rerenders
  if (!ref.current.order) {
    atom._i = atom._i ? atom._i + 1 : 1
    ref.current.order = atom._i
  }

  // cancel any pending scheduled updates after each render
  // since we just go rerendered by the parent component
  if (ref.current.cancelUpdate) {
    ref.current.cancelUpdate()
    ref.current.cancelUpdate = null
  }

  useEffect(function () {
    if (!observe) { return }
    ref.current.unobserve = atom.observe(onChange, ref.current.order)
    onChange()
    function onChange () {
      // store updates happening in rapid sequence
      // get cancelled and rescheduled
      if (ref.current.cancelUpdate) {
        ref.current.cancelUpdate()
        ref.current.cancelUpdate = null
      }
      // schedule an update
      ref.current.cancelUpdate = schedule(function () {
        ref.current.cancelUpdate = null
        var nextMappedProps = selector(atom.get())
        if (!pure || differ(ref.current.mappedProps, nextMappedProps)) {
          triggerRerender({})
        }
      })()
    }
    return function () {
      ref.current.unobserve && ref.current.unobserve()
      if (ref.current.cancelUpdate) {
        ref.current.cancelUpdate()
        ref.current.cancelUpdate = null
      }
    }
  }, [atom])

  // always return fresh mapped props, in case
  // this is a parent rerendering children
  return ref.current.mappedProps
}

function useActions () {
  var ref = useContext(AtomContext);
  var atom = ref.atom;
  return atom.actions
}

function useDispatch () {
  var ref = useContext(AtomContext);
  var atom = ref.atom;
  return atom.dispatch
}

function assert (cond, error) {
  if (!cond) {
    throw new Error(error)
  }
}

module.exports = { useAtom: useAtom, useActions: useActions, useDispatch: useDispatch }
